import asyncio
import os
from pathlib import Path
import re
import time
from httpx import AsyncClient
import requests
import json
import argparse

from bilix.sites.bilibili import api
from rich import print


def arg_parse():
    parser = argparse.ArgumentParser()
    
    ranking_group = parser.add_argument_group()
    ranking_group.title = 'by ranking'
    ranking_group.add_argument('--by-ranking', action='store_true', help='从排行榜获取 bvids')
    ranking_group.add_argument('--ranking-rid', type=int, default=0, help='目标排行 rid，0 为全站排行榜 [default: 0]')

    up_videos_group = parser.add_argument_group()
    up_videos_group.title = 'by up videos'
    up_videos_group.add_argument('--by-up_videos', action='store_true', help='从 up 主用户页获取全部的投稿的 bvids')
    up_videos_group.add_argument('--up_videos-mid', type=str, help='目标 up 主的 mid (也可以是用户页的 URL)')

    args = parser.parse_args()
    return args

def by_raning(rid: int) -> Path:
    bilibili_ranking_api = "https://api.bilibili.com/x/web-interface/ranking/v2"
    bilibili_ranking_params = {
        "rid": rid,
        "type": "all"
    }

    r = requests.get(bilibili_ranking_api, params=bilibili_ranking_params)
    r.raise_for_status()
    ranking_json = json.loads(r.text)
    assert ranking_json['code'] == 0 # 0 为成功（HTTP 200 不能信）

    ranking = ranking_json['data']['list']
    bvids = []
    for video_info in ranking:
        # print(video_info['title'], video_info['bvid'], video_info['pic'])
        bvid = video_info['bvid']
        bvids.append(bvid)

    import datetime
    today = datetime.date.today()
    os.makedirs('bvids', exist_ok=True)

    bvids_filepath = f'bvids/by-ranking/rid-{rid}/rid-{rid}-{int(time.time())}.txt'
    os.makedirs(os.path.dirname(bvids_filepath), exist_ok=True)
    with open(bvids_filepath, 'w', encoding='utf-8') as f:
        for bvid in bvids:
            f.write(f'{bvid}' + '\n')
    abs_filepath = os.path.abspath(bvids_filepath)
    print(f'已保存 {len(bvids)} 个 bvid 到 {abs_filepath}')
    return Path(abs_filepath)

async def by_up_videos(url_or_mid: str) -> Path:
    ''' 频率高了会封 '''

    if isinstance(url_or_mid, int):
        mid = str(url_or_mid)
    elif url_or_mid.startswith("http"):
        mid = re.findall(r"/(\d+)", url_or_mid)[0]
    else:
        mid = url_or_mid

    assert isinstance(mid, str)
    assert mid.isdigit(), 'mid 应是数字字符串'

    client = AsyncClient(**api.dft_client_settings)
    ps = 30 # 每页视频数，最小 1，最大 50，默认 30
    order = 'pubdate'# 默认为pubdate 最新发布：pubdate 最多播放：click 最多收藏：stow
    keyword = '' # 搜索关键词
    bv_ids = []
    pn = 1
    up_name, total_size, bv_ids_page = await api.get_up_info(client, mid, pn, ps, order, keyword)
    bv_ids += bv_ids_page
    print(f'{mid} {up_name} 共 {total_size} 个视频... （第 1 页）')
    while pn < total_size / ps:
        pn += 1
        print(f'获取第 {pn} 页 (10s...)')
        await asyncio.sleep(10)
        _, _, bv_ids_page = await api.get_up_info(client, mid, pn, ps, order, keyword)
        bv_ids += bv_ids_page

    print(mid, up_name, total_size)
    await client.aclose()
    assert len(bv_ids) == len(set(bv_ids)), '有重复的 bv_id'
    assert total_size == len(bv_ids), '视频总数不匹配'
    filepath = f'bvids/by-up_videos/mid-{mid}-{int(time.time())}.txt'
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    abs_filepath = os.path.abspath(filepath)
    with open(abs_filepath, 'w', encoding='utf-8') as f:
        for bv_id in bv_ids:
            f.write(f'{bv_id}' + '\n')
    print(f'已保存 {len(bv_ids)} 个 bvid 到 {abs_filepath}')
    return Path(abs_filepath)


async def _main():
    args = arg_parse()
    if args.by_ranking:
        by_raning(args.ranking_rid)
    if args.by_up_videos:
        assert args.up_videos_mid is not None, 'up_videos_mid 不能为空'
        await by_up_videos(args.up_videos_mid)

def main():
    asyncio.run(_main())

if __name__ == '__main__':
    main()


import os
import argparse
from dataclasses import dataclass
from pathlib import Path

from biliarchiver._biliarchiver_upload_bvid import upload_bvid
from biliarchiver.config import config

@dataclass
class Args:
    bvids: str
    by_storage_home_dir: bool


def parse_args():
    parser = argparse.ArgumentParser()
    bvids_file_group = parser.add_argument_group()
    bvids_file_group.title = 'bvids'
    bvids_file_group.add_argument('--bvids', type=str, dest='bvids',
        help='bvids 列表的文件路径')
    storage_home_dir_group = parser.add_argument_group()
    storage_home_dir_group.title = 'storage_home_dir'
    storage_home_dir_group.add_argument('--by-storage_home_dir', action='store_true', dest='by_storage_home_dir',
        help='从 config.json 中读取 storage_home_dir，然后上传 storage_home_dir/videos 下的所有视频')

    args = Args(**vars(parser.parse_args()))

    return args

def main():
    args = parse_args()
    if args.by_storage_home_dir:
        for bvid_with_upper_part in os.listdir(config.storage_home_dir / 'videos'):
            bvid = bvid_with_upper_part
            if '-' in bvid_with_upper_part:
                bvid = bvid_with_upper_part.split('-')[0]

            upload_bvid(bvid)
    
    if args.bvids:
        with open(args.bvids, 'r', encoding='utf-8') as f:
            bvids_from_file = f.read().splitlines()
        for bvid in bvids_from_file:

            upload_bvid(bvid)

if __name__ == '__main__':
    main()


from dataclasses import dataclass
import os
import json
from pathlib import Path

CONFIG_FILE = 'config.json'
BILIBILI_IDENTIFIER_PERFIX = 'BiliBili' # IA identifier 前缀。


class singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass
class _Config(metaclass=singleton):
    video_concurrency: int = 3 
    part_concurrency: int = 10
    stream_retry: int = 20
    storage_home_dir: Path = Path('bilibili_archive_dir/').expanduser()
    ia_key_file: Path = Path('~/.bili_ia_keys.txt').expanduser()
    cookies_file: Path = Path('~/.cookies.txt').expanduser()

    def __init__(self):
        self.is_right_pwd()
        if not os.path.exists(CONFIG_FILE):
            print(f'{CONFIG_FILE} 不存在，创建中...')
            self.save()
        with open(CONFIG_FILE, 'r', encoding='utf-8') as f:
            print(f'Loading {CONFIG_FILE} ...')
            config_file = json.load(f)

        self.video_concurrency: int = config_file['video_concurrency']
        self.part_concurrency: int = config_file['part_concurrency']
        self.stream_retry: int = config_file['stream_retry']

        self.storage_home_dir: Path = Path(config_file['storage_home_dir']).expanduser()
        self.ia_key_file: Path = Path(config_file['ia_key_file']).expanduser()

    def save(self):
        with open(CONFIG_FILE, 'w', encoding='utf-8') as f:
            json.dump({
                'video_concurrency': self.video_concurrency,
                'part_concurrency': self.part_concurrency,
                'stream_retry': self.stream_retry,
                'storage_home_dir': str(self.storage_home_dir),
                'ia_key_file': str(self.ia_key_file),
                'cookies_file': str(self.cookies_file),
            }, f, ensure_ascii=False, indent=4)

    def is_right_pwd(self):
        if not os.path.exists('biliarchiver.home'):
            raise Exception('先在当前工作目录创建 biliarchiver.home 文件')

config = _Config()



import asyncio
import os
import argparse
from pathlib import Path
from typing import Union

from internetarchive import get_item

from biliarchiver.archive_bvid import archive_bvid
from biliarchiver.config import config

from bilix.sites.bilibili.downloader import DownloaderBilibili
from rich.console import Console
from httpx import AsyncClient, Client
from rich.traceback import install
install()

from biliarchiver.utils.string import human_readable_upper_part_map
from biliarchiver.utils.ffmpeg import check_ffmpeg

from biliarchiver.config import BILIBILI_IDENTIFIER_PERFIX

from dataclasses import dataclass

@dataclass
class Args:
    bvids: str
    skip_ia: bool

def parse_args():

    parser = argparse.ArgumentParser()
    parser.add_argument('--bvids', dest='bvids', type=str, help='bvids 列表的文件路径', required=True)
    parser.add_argument('-s', '--skip-ia-check', dest='skip_ia', action='store_true',
                        help='不检查 IA 上是否已存在对应 BVID 的 item ，直接开始下载')
    
    args = Args(**vars(parser.parse_args()))

    return args

def check_ia_item_exist(client: Client, identifier: str) -> bool:
    cache_dir = config.storage_home_dir / 'ia_item_exist_cache'
    cache_dir.mkdir(parents=True, exist_ok=True)
    def create_item_exist_cache_file(identifier: str) -> Path:
        with open(cache_dir / f'{identifier}.mark', 'w', encoding='utf-8') as f:
            f.write('')
        return cache_dir / f'{identifier}.mark'
    def check_ia_item_exist_from_cache_file(identifier: str) -> bool:
        return (cache_dir / f'{identifier}.mark').exists()


    if check_ia_item_exist_from_cache_file(identifier):
        return True

    # params = {
    #     'identifier': identifier,
    #     'output': 'json',
    # }
    # r = client.get('https://archive.org/services/check_identifier.php' ,params=params)
    # r.raise_for_status()
    # r_json = r.json()
    # assert r_json['type'] =='success'
    # if r_json['code'] == 'available':
    #     return False
    # elif r_json['code'] == 'not_available':
    #     return True
    # else:
    #     raise ValueError(f'Unexpected code: {r_json["code"]}')
    item = get_item(identifier)
    if item.exists:
        create_item_exist_cache_file(identifier)
        return True

    return False

def _main():
    args = parse_args()
    assert check_ffmpeg() is True, 'ffmpeg 未安装'

    assert args.bvids is not None, '必须指定 bvids 列表的文件路径'
    with open(args.bvids, 'r', encoding='utf-8') as f:
        bvids_from_file = f.read().splitlines()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    d = DownloaderBilibili(hierarchy=True, sess_data=None, # sess_data 将在后面装载 cookies 时装载 # type: ignore
        video_concurrency=config.video_concurrency,
        part_concurrency=config.part_concurrency,
        stream_retry=config.stream_retry,
    )
    update_cookies_from_file(d.client, config.cookies_file)
    client = Client(cookies=d.client.cookies, headers=d.client.headers)
    logined = is_login(client)
    if not logined:
        return

    d.progress.start()
    for index, bvid in enumerate(bvids_from_file):
        if not args.skip_ia:
            upper_part = human_readable_upper_part_map(string=bvid, backward=True)
            remote_identifier = f'{BILIBILI_IDENTIFIER_PERFIX}-{bvid}_p1-{upper_part}'
            if check_ia_item_exist(client, remote_identifier):
                print(f'IA 上已存在 {remote_identifier} ，跳过')
                continue

        while len(asyncio.all_tasks(loop)) > config.video_concurrency:
            loop.run_until_complete(asyncio.sleep(0.008))

        print(f'=== {bvid} ({index+1}/{len(bvids_from_file)}) ===')

        task = loop.create_task(archive_bvid(d, bvid, logined=logined))
    
    while len(asyncio.all_tasks(loop)) > 0:
        loop.run_until_complete(asyncio.sleep(1))
    


def update_cookies_from_file(client: AsyncClient, cookies_path: Union[str, Path]):
    if isinstance(cookies_path, Path):
        cookies_path = cookies_path.expanduser()
    elif isinstance(cookies_path, str):
        cookies_path = Path(cookies_path).expanduser()
    else:
        raise TypeError(f'cookies_path: {type(cookies_path)}')

    assert os.path.exists(cookies_path), f'cookies 文件不存在: {cookies_path}'
    from http.cookiejar import MozillaCookieJar
    cj = MozillaCookieJar()
    cj.load(f'{cookies_path}', ignore_discard=True, ignore_expires=True)
    loadded_cookies = 0
    for cookie in cj:
        # only load bilibili cookies
        if 'bilibili' in cookie.domain:
            assert cookie.value is not None
            client.cookies.set(
                cookie.name, cookie.value, domain=cookie.domain, path=cookie.path
                )
            loadded_cookies += 1
    print(f'从 {cookies_path} 品尝了 {loadded_cookies} 块 cookies')
    if loadded_cookies > 100:
        print('吃了过多的 cookies，可能导致 httpx.Client 怠工，响应非常缓慢')

    assert client.cookies.get('SESSDATA') is not None, 'SESSDATA 不存在'
    # print(f'SESS_DATA: {client.cookies.get("SESSDATA")}')

def is_login(cilent: Client) -> bool:
    r = cilent.get('https://api.bilibili.com/x/member/web/account')
    r.raise_for_status()
    nav_json = r.json()
    if nav_json['code'] == 0:
        print('BiliBili 登录成功，饼干真香。')
        print('NOTICE: 存档过程中请不要在 cookies 的源浏览器访问 B 站，避免 B 站刷新'
              ' cookies 导致我们半路全下到的视频全是 480P 的优酷土豆级醇享画质。')
        return True
    print('未登录/SESSDATA无效/过期，你这饼干它保真吗？')
    return False

def main():
    try:
        _main()
    except KeyboardInterrupt:
        print('KeyboardInterrupt')
    finally:
        # 显示终端光标
        console = Console()
        console.show_cursor()

if __name__ == '__main__':
    main()


import asyncio
import os
from pathlib import Path

import aiofiles
import httpx
from bilix.download.utils import raise_api_error, req_retry
from bilix.exception import APIError

from bilix.sites.bilibili import api

from rich import print
import json

from bilix.sites.bilibili.downloader import DownloaderBilibili
from biliarchiver.config import BILIBILI_IDENTIFIER_PERFIX
from biliarchiver.config import config
from biliarchiver.utils.string import human_readable_upper_part_map

@raise_api_error
async def new_get_subtitle_info(client: httpx.AsyncClient, bvid, cid):
    params = {'bvid': bvid, 'cid': cid}
    res = await req_retry(client, 'https://api.bilibili.com/x/player/v2', params=params)
    info = json.loads(res.text)
    if info['code'] == -400:
        raise APIError(f'未找到字幕信息', params)

    # 这里 monkey patch 一下把返回 lan_doc 改成返回 lan，这样生成的字幕文件名就是 语言代码 而不是 中文名 了
    # 例如
    # lan_doc: 中文（中国）
    # lang: zh-CN

  # return [[f'http:{i["subtitle_url"]}', i['lan_doc']] for i in info['data']['subtitle']['subtitles']]
    return [[f'http:{i["subtitle_url"]}', i['lan']] for i in info['data']['subtitle']['subtitles']]
api.get_subtitle_info = new_get_subtitle_info


async def archive_bvid(d: DownloaderBilibili, bvid: str, logined: bool=False):
    assert d.hierarchy is True, 'hierarchy 必须为 True' # 为保持后续目录结构、文件命名的一致性
    assert d.client.cookies.get('SESSDATA') is not None, 'sess_data 不能为空' # 开个大会员呗，能下 4k 呢。
    assert logined is True, '请先检查 SESSDATA 是否过期，再将 logined 设置为 True' # 防误操作

    upper_part = human_readable_upper_part_map(string=bvid, backward=True)
    OLD_videos_basepath: Path = config.storage_home_dir / 'videos' / bvid
    videos_basepath: Path = config.storage_home_dir / 'videos' / f'{bvid}-{upper_part}'

    if os.path.exists(OLD_videos_basepath):
        print(f'检测到旧的视频目录 {OLD_videos_basepath}，将其重命名为 {videos_basepath}...')
        os.rename(OLD_videos_basepath, videos_basepath)


    if os.path.exists(videos_basepath / '_all_downloaded.mark'):
        print(f'{bvid} 所有分p都已下载过了')
        return

    url = f'https://www.bilibili.com/video/{bvid}/'
    # 为了获取 pages，先请求一次
    first_video_info = await api.get_video_info(d.client, url)

    os.makedirs(videos_basepath, exist_ok=True)

    pid = 0
    for page in first_video_info.pages:
        pid += 1 # pid 从 1 开始
        if not page.p_url.endswith(f'?p={pid}'):
            print(f'{bvid} 的 P{pid} 不存在 (可能视频被 UP主/B站 删了)')
            continue

        file_basename = f'{bvid}_p{pid}'
        video_basepath = videos_basepath / f'{BILIBILI_IDENTIFIER_PERFIX}-{file_basename}'
        video_extrapath = video_basepath / 'extra'
        if os.path.exists(f'{video_basepath}/_downloaded.mark'):
            print(f'{file_basename}: 已经下载过了')
            continue

        video_info = await api.get_video_info(d.client, page.p_url)
        print(f'{file_basename}: {video_info.title}...')
        os.makedirs(video_basepath, exist_ok=True)
        os.makedirs(video_extrapath, exist_ok=True)


        old_p_name = video_info.pages[video_info.p].p_name
        old_h1_title = video_info.h1_title
    
        # 在 d.hierarchy is True 且 h1_title 超长的情况下， bilix 会将 p_name 作为文件名
        video_info.pages[video_info.p].p_name = file_basename # 所以这里覆盖 p_name 为 file_basename
        video_info.h1_title = 'iiiiii' * 50 # 然后假装超长标题
        # 这样 bilix 保存的文件名就是我们想要的了（谁叫 bilix 不支持自定义文件名呢）
        # NOTE: p_name 似乎也不宜过长，否则还是会被 bilix 截断。
        # 但是我们以 {bvid}_p{pid} 作为文件名，这个长度是没问题的。


        # 选择编码，优先 hevc，没有的话就 avc
        # 不选 av0 ，毕竟目前没几个设备能拖得动
        codec = None
        if video_info.dash:
            for media in video_info.dash.videos:
                if media.codec.startswith('hev'):
                    codec = media.codec
                    print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                    break
            if codec is None:
                for media in video_info.dash.videos:
                    if media.codec.startswith('avc'):
                        codec = media.codec
                        print(f'{file_basename}: "{codec}" "{media.quality}" ...')
                        break
            assert codec is not None, f'{file_basename}: 没有 avc 或 hevc 编码的视频'
        elif video_info.other:
            print(f'{file_basename}: 未解析到dash资源，交给 bilix 处理 ...')
            codec = ''
        else:
            raise APIError(f'{file_basename}: 未解析到视频资源', page.p_url)

        assert codec is not None

        cor1 = d.get_video(page.p_url ,video_info=video_info, path=video_basepath,
                    quality=0, # 选择最高画质
                    codec=codec, # 编码
                    # 下载 ass 弹幕(bilix 会自动调用 danmukuC 将 pb 弹幕转为 ass)、封面、字幕
                    # 弹幕、封面、字幕都会被放进 extra 子目录里，所以需要 d.hierarchy is True
                    dm=True, image=True, subtitle=True
                    )
        # 下载原始的 pb 弹幕
        cor2 = d.get_dm(page.p_url, video_info=video_info, path=video_extrapath)
        # 下载视频超详细信息（BV 级别，不是分 P 级别）
        cor3 = download_bilibili_video_detail(d.client, bvid, f'{video_extrapath}/{file_basename}.info.json')
        await asyncio.gather(cor1, cor2, cor3)

        # 还原为了自定义文件名而做的覆盖
        video_info.pages[video_info.p].p_name = old_p_name
        video_info.h1_title = old_h1_title

        # 单 p 下好了
        async with aiofiles.open(f'{video_basepath}/_downloaded.mark', 'w', encoding='utf-8') as f:
            await f.write('')


    # bv 对应的全部 p 下好了
    async with aiofiles.open(f'{videos_basepath}/_all_downloaded.mark', 'w', encoding='utf-8') as f:
            await f.write('')

    


async def download_bilibili_video_detail(client, bvid, filename):
    if os.path.exists(filename):
        print(f'{bvid} 视频详情已存在')
        return
    # url = 'https://api.bilibili.com/x/web-interface/view'
    url = 'https://api.bilibili.com/x/web-interface/view/detail' # 超详细 API（BV 级别，不是分 P 级别）
    params = {'bvid': bvid}
    r = await req_retry(client, url, params=params ,follow_redirects=True)
    r.raise_for_status()
    r_json = r.json()
    assert r_json['code'] == 0, f'{bvid} 视频详情获取失败'

    async with aiofiles.open(filename, 'w', encoding='utf-8') as f:
        # f.write(json.dumps(r.json(), indent=4, ensure_ascii=False))
        await f.write(r.text)
    print(f'{bvid} 视频详情已保存')
